# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import warnings
from datetime import datetime, date
from dateutil.parser import parse
from ..config import ERROR


# Module API

def cast_date(format, value, **options):
    if isinstance(value, datetime):
        value_time = value.time()
        if value_time.hour == 0 and value_time.minute == 0 and value_time.second == 0:
            return datetime(value.year, value.month, value.day).date()
        else:
            return ERROR

    if isinstance(value, date):
        return value

    if not isinstance(value, six.string_types):
        return ERROR

    # Parse string date
    try:
        if format == 'default':
            value = datetime.strptime(value, _DEFAULT_PATTERN).date()
        elif format == 'any':
            value = parse(value).date()
        else:
            if format.startswith('fmt:'):
                warnings.warn(
                    'Format "fmt:<PATTERN>" is deprecated. '
                    'Please use "<PATTERN>" without "fmt:" prefix.',
                    UserWarning)
                format = format.replace('fmt:', '')
            value = datetime.strptime(value, format).date()
    except Exception:
        return ERROR

    return value


# Internal

_DEFAULT_PATTERN = '%Y-%m-%d'



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import six
import uuid
import base64
import rfc3986.exceptions
import rfc3986.validators
import rfc3986.uri
from ..config import ERROR


# Module API

def cast_string(format, value, **options):
    if not isinstance(value, six.string_types):
        return ERROR
    if format == 'uri':
        uri = _uri_from_string(value)
        try:
            _uri_validator.validate(uri)
        except rfc3986.exceptions.ValidationError:
            return ERROR
    elif format == 'email':
        if not re.match(_EMAIL_PATTERN, value):
            return ERROR
    elif format == 'uuid':
        try:
            uuid.UUID(value, version=4)
        except Exception:
            return ERROR
    elif format == 'binary':
        try:
            base64.b64decode(value)
        except Exception:
            return ERROR
    return value


# Internal

_EMAIL_PATTERN = re.compile(r'[^@]+@[^@]+\.[^@]+')
_uri_from_string = rfc3986.uri.URIReference.from_string
_uri_validator = rfc3986.validators.Validator().require_presence_of('scheme')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
import datetime
import isodate
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', isodate.Duration(years=1), isodate.Duration(years=1)),
    ('default', 'P1Y10M3DT5H11M7S',
         isodate.Duration(years=1, months=10, days=3, hours=5, minutes=11, seconds=7)),
    ('default', 'P1Y', isodate.Duration(years=1)),
    ('default', 'P1M', isodate.Duration(months=1)),
    ('default', 'PT1S', datetime.timedelta(seconds=1)),
    ('default', datetime.timedelta(seconds=1),  datetime.timedelta(seconds=1)),
    ('default', 'P1M1Y', ERROR),
    ('default', 'P-1Y', ERROR),
    ('default', 'year', ERROR),
    ('default', True, ERROR),
    ('default', False, ERROR),
    ('default', 1, ERROR),
    ('default', '', ERROR),
    ('default', [], ERROR),
    ('default', {}, ERROR),
])
def test_cast_duration(format, value, result):
    assert types.cast_duration(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
COMPILED_RE = type(re.compile(""))


# Module API

def check_pattern(constraint, value):
    if value is None:
        return True
    if not isinstance(constraint, COMPILED_RE):
        regex = re.compile('^{0}$'.format(constraint))
    else:
        regex = constraint
    match = regex.match(value)
    if match:
        return True
    return False



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import six
import json
from collections import OrderedDict
from copy import deepcopy
from six.moves import zip_longest
from .profile import Profile
from .field import Field
from . import exceptions
from . import helpers
from . import config
from . import types


# Module API

class Schema(object):
    """Schema representation

    # Arguments
        descriptor (str/dict): schema descriptor one of:
            - local path
            - remote url
            - dictionary
        strict (bool): flag to specify validation behaviour:
            - if false, errors will not be raised but instead collected in `schema.errors`
            - if true, validation errors are raised immediately

    # Raises
        TableSchemaException: raise any error that occurs during the process

    """

    # Public

    def __init__(self, descriptor={}, strict=False):

        # Process descriptor
        descriptor = helpers.retrieve_descriptor(descriptor)

        # Set attributes
        self.__strict = strict
        self.__current_descriptor = deepcopy(descriptor)
        self.__next_descriptor = deepcopy(descriptor)
        self.__profile = Profile('table-schema')
        self.__errors = []
        self.__fields = []

        # Build instance
        self.__build()

    @property
    def valid(self):
        """Validation status

        Always true in strict mode.

        # Returns
            bool: validation status

        """
        return not bool(self.__errors)

    @property
    def errors(self):
        """Validation errors

        Always empty in strict mode.

        # Returns
            Exception[]: validation errors

        """
        return self.__errors

    @property
    def descriptor(self):
        """Schema's descriptor

        # Returns
            dict: descriptor

        """
        # Never use this.descriptor inside this class (!!!)
        return self.__next_descriptor

    @property
    def missing_values(self):
        """Schema's missing values

        # Returns
            str[]: missing values

        """
        return self.__current_descriptor.get('missingValues', [])

    @property
    def primary_key(self):
        """Schema's primary keys

        # Returns
            str[]: primary keys

        """
        primary_key = self.__current_descriptor.get('primaryKey', [])
        if not isinstance(primary_key, list):
            primary_key = [primary_key]
        return primary_key

    @property
    def foreign_keys(self):
        """Schema's foreign keys

        # Returns
            dict[]: foreign keys

        """
        foreign_keys = self.__current_descriptor.get('foreignKeys', [])
        for key in foreign_keys:
            key.setdefault('fields', [])
            key.setdefault('reference', {})
            key['reference'].setdefault('resource', '')
            key['reference'].setdefault('fields', [])
            if not isinstance(key['fields'], list):
                key['fields'] = [key['fields']]
            if not isinstance(key['reference']['fields'], list):
                key['reference']['fields'] = [key['reference']['fields']]
        return foreign_keys

    @property
    def fields(self):
        """Schema's fields

        # Returns
            Field[]: an array of field instances

        """
        return self.__fields

    @property
    def field_names(self):
        """Schema's field names

        # Returns
            str[]: an array of field names

        """
        return [field.name for field in self.fields]

    def get_field(self, name):
        """Get schema's field by name.

        > Use `table.update_field` if you want to modify the field descriptor

        # Arguments
            name (str): schema field name

        # Returns
           Field/None: `Field` instance or `None` if not found

        """
        for field in self.fields:
            if field.name == name:
                return field
        return None

    def add_field(self, descriptor):
        """ Add new field to schema.

        The schema descriptor will be validated with newly added field descriptor.

        # Arguments
            descriptor (dict): field descriptor

        # Raises
            TableSchemaException: raises any error that occurs during the process

        # Returns
            Field/None: added `Field` instance or `None` if not added

        """
        self.__current_descriptor.setdefault('fields', [])
        self.__current_descriptor['fields'].append(descriptor)
        self.__build()
        return self.__fields[-1]

    def update_field(self, name, update):
        """Update existing descriptor field by name

        # Arguments
            name (str): schema field name
            update (dict): update to apply to field's descriptor

        # Returns
            bool: true on success and false if no field is found to be modified

        """
        for field in self.__next_descriptor['fields']:
            if field['name'] == name:
                field.update(update)
                return True
        return False

    def remove_field(self, name):
        """Remove field resource by name.

        The schema descriptor will be validated after field descriptor removal.

        # Arguments
            name (str): schema field name

        # Raises
            TableSchemaException: raises any error that occurs during the process

        # Returns
            Field/None: removed `Field` instances or `None` if not found

        """
        field = self.get_field(name)
        if field:
            predicat = lambda field: field.get('name') != name
            self.__current_descriptor['fields'] = list(filter(
                predicat, self.__current_descriptor['fields']))
            self.__build()
        return field

    def cast_row(self, row, fail_fast=False, row_number=None, exc_handler=None):
        """Cast row based on field types and formats.

        # Arguments
            row (any[]: data row as an array of values

        # Returns
            any[]: returns cast data row

        """
        exc_handler = helpers.default_exc_handler if exc_handler is None else \
            exc_handler

        # Prepare
        result = []
        errors = []
        if row_number is not None:
            row_number_info = ' for row "%s"' % row_number
        else:
            row_number_info = ''
        # Check row length
        if len(row) != len(self.fields):
            message = (
                'Row length %s doesn\'t match fields count %s' +
                row_number_info) % (len(row), len(self.fields))
            exc = exceptions.CastError(message)
            # Some preparations for error reporting, relevant if custom error
            # handling is in place.
            if len(row) < len(self.fields):
                # Treat missing col values as None
                keyed_row = OrderedDict(
                    zip_longest((field.name for field in self.fields), row))
                # Use added None values for further processing
                row = list(keyed_row.values())
            else:
                fields = self.fields
                keyed_row = OrderedDict(
                    # Use extra column number if value index exceeds fields
                    (fields[i].name if fields[i:]
                     else 'tableschema-cast-error-extra-col-{}'.format(i+1),
                     value)
                    for (i, value) in enumerate(row))
            exc_handler(exc, row_number=row_number, row_data=keyed_row,
                        error_data=keyed_row)

        # Cast row
        for field, value in zip(self.fields, row):
            try:
                result.append(field.cast_value(value))
            except exceptions.CastError as exception:
                if fail_fast:
                    raise
                # Wrap original value in a FailedCast object to be able to
                # further process/yield values and to distinguish uncasted
                # values on the consuming side.
                result.append(FailedCast(value))
                errors.append(exception)

        # Raise errors
        if errors:
            message = (
                'There are %s cast errors (see exception.errors)' +
                row_number_info) % len(errors)
            keyed_row = OrderedDict(zip(self.field_names, row))
            # Add the cast failure-causing fields only to error data.
            # Indexing results with the row field index should be ok at this
            # point due to the previous processing.
            error_data = OrderedDict(
                (name, value)
                for (i, (name, value)) in enumerate(keyed_row.items())
                if isinstance(result[i], FailedCast))
            exc_handler(
                exceptions.CastError(message, errors=errors),
                row_number=row_number, row_data=keyed_row,
                error_data=error_data)

        return result

    def infer(self, rows, headers=1, confidence=0.75,
              guesser_cls=None, resolver_cls=None):
        """Infer and set `schema.descriptor` based on data sample.

        # Arguments
            rows (list[]): array of arrays representing rows.
            headers (int/str[]): data sample headers (one of):
              - row number containing headers (`rows` should contain headers rows)
              - array of headers (`rows` should NOT contain headers rows)
            confidence (float): how many casting errors are allowed (as a ratio, between 0 and 1)
            guesser_cls (class): you can implement inferring strategies by
                 providing type-guessing and type-resolving classes [experimental]
            resolver_cls (class): you can implement inferring strategies by
                 providing type-guessing and type-resolving classes [experimental]

        # Returns
            dict: Table Schema descriptor

        """

        # Get headers
        if isinstance(headers, int):
            headers_row = headers
            while True:
                headers_row -= 1
                headers = rows.pop(0)
                if not headers_row:
                    break
        elif isinstance(headers, list):
            seen_cells = []
            headers = list(headers)
            for index, cell in enumerate(headers):
                count = seen_cells.count(cell) + 1
                headers[index] = '%s%s' % (cell, count) if count > 1 else cell
                seen_cells.append(cell)
        elif not isinstance(headers, list):
            headers = []

        # Get descriptor
        missing_values = self.__current_descriptor.get('missingValues', config.DEFAULT_MISSING_VALUES)
        guesser = guesser_cls() if guesser_cls else _TypeGuesser(missing_values)
        resolver = (resolver_cls or _TypeResolver)()
        descriptor = {'fields': [], 'missingValues': missing_values}
        type_matches = {}
        for number, header in enumerate(headers, start=1):
            descriptor['fields'].append({'name': header or 'field%s' % number})
        for index, row in enumerate(rows):
            # Normalize rows with invalid dimensions for sanity
            row_length = len(row)
            headers_length = len(headers)
            if row_length > headers_length:
                row = row[:len(headers)]
            if row_length < headers_length:
                diff = headers_length - row_length
                fill = [''] * diff
                row = row + fill
            # build a column-wise lookup of type matches
            for index, value in enumerate(row):
                rv = guesser.cast(value)
                if type_matches.get(index):
                    type_matches[index].extend(rv)
                else:
                    type_matches[index] = list(rv)
        # choose a type/format for each column based on the matches
        for index, results in type_matches.items():
            rv = resolver.get(results, confidence)
            descriptor['fields'][index].update(**rv)

        # Save descriptor
        self.__current_descriptor = descriptor
        self.__build()

        return descriptor

    def commit(self, strict=None):
        """Update schema instance if there are in-place changes in the descriptor.

        # Example

        ```python
        from tableschema import Schema
        descriptor = {'fields': [{'name': 'my_field', 'title': 'My Field', 'type': 'string'}]}
        schema = Schema(descriptor)
        print(schema.get_field('my_field').descriptor['type']) # string

        # Update descriptor by field position
        schema.descriptor['fields'][0]['type'] = 'number'
        # Update descriptor by field name
        schema.update_field('my_field', {'title': 'My Pretty Field'}) # True

        # Change are not committed
        print(schema.get_field('my_field').descriptor['type']) # string
        print(schema.get_field('my_field').descriptor['title']) # My Field

        # Commit change
        schema.commit()
        print(schema.get_field('my_field').descriptor['type']) # number
        print(schema.get_field('my_field').descriptor['title']) # My Pretty Field

        ```

        # Arguments
            strict (bool): alter `strict` mode for further work

        # Raises
            TableSchemaException: raises any error that occurs during the process

        # Returns
            bool: true on success and false if not modified

        """
        if strict is not None:
            self.__strict = strict
        elif self.__current_descriptor == self.__next_descriptor:
            return False
        self.__current_descriptor = deepcopy(self.__next_descriptor)
        self.__build()
        return True

    def save(self, target, ensure_ascii=True):
        """Save schema descriptor to target destination.

        # Arguments
            target (str): path where to save a descriptor

        # Raises
            TableSchemaException: raises any error that occurs during the process

        # Returns
            bool: true on success

        """
        mode = 'w'
        encoding = 'utf-8'
        if six.PY2:
            mode = 'wb'
            encoding = None
        helpers.ensure_dir(target)
        with io.open(target, mode=mode, encoding=encoding) as file:
            json.dump(self.__current_descriptor, file, indent=4, ensure_ascii=ensure_ascii)

    # Internal

    def __build(self):

        # Process descriptor
        expand = helpers.expand_schema_descriptor
        self.__current_descriptor = expand(self.__current_descriptor)
        self.__next_descriptor = deepcopy(self.__current_descriptor)

        # Validate descriptor
        try:
            self.__profile.validate(self.__current_descriptor)
            self.__errors = []
        except exceptions.ValidationError as exception:
            self.__errors = exception.errors
            if self.__strict:
                raise exception

        # Populate fields
        self.__fields = []
        for field in self.__current_descriptor.get('fields', []):
            missing_values = self.__current_descriptor['missingValues']
            try:
                field = Field(field, missing_values=missing_values, schema=self)
            except exceptions.TableSchemaException as e:
                if self.__strict:
                    raise e
                else:
                    field = False
            self.__fields.append(field)

    # Deprecated

    headers = field_names
    has_field = get_field


class FailedCast(object):
    """Wrap an original data field value that failed to be properly casted.

    FailedCast allows for further processing/yielding values but still be able
    to distinguish uncasted values on the consuming side.

    Delegates attribute access and the basic rich comparison methods to the
    underlying object. Supports default user-defined classes hashability i.e.
    is hashable based on object identity (not based on the wrapped value).

    # Arguments
        value (any): value

    """

    # Make this "reasonably immutable": Don't support setting other attributes,
    # don't support modifying re-setting value
    __slots__ = ('_value',)

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return 'FailedCast(%r)' % self._value

    def __getattr__(self, name):
        return getattr(self._value, name)

    def __lt__(self, other):
        return self._value < other

    def __le__(self, other):
        return self._value <= other

    def __eq__(self, other):
        return self._value == other

    def __ne__(self, other):
        return self._value != other

    def __gt__(self, other):
        return self._value > other

    def __ge__(self, other):
        return self._value >= other

    def __hash__(self):
        return object.__hash__(self)


# Internal

_INFER_TYPE_ORDER = [
    'duration',
    'geojson',
    'geopoint',
    'object',
    'array',
    'datetime',
    'time',
    'date',
    'integer',
    'number',
    'boolean',
    'string',
    'any',
]


class _TypeGuesser(object):
    """Guess the type for a value returning a tuple of ('type', 'format')
    """

    # Public

    def __init__(self, missing_values):
        self.missing_values = missing_values

    def cast(self, value):
        for priority, name in enumerate(_INFER_TYPE_ORDER):
            cast = getattr(types, 'cast_%s' % name)
            if value not in self.missing_values:
                result = cast('default', value)
                if result != config.ERROR:
                    yield (name, 'default', priority)


class _TypeResolver(object):
    """Get the best matching type/format from a list of possible ones.
    """

    # Public

    def get(self, results, confidence):
        variants = set(results)
        # only one candidate... that's easy.
        if len(variants) == 1:
            rv = {'type': results[0][0], 'format': results[0][1]}
        else:
            counts = {}
            for result in results:
                if counts.get(result):
                    counts[result] += 1
                else:
                    counts[result] = 1
            # tuple representation of `counts` dict sorted by values
            sorted_counts = sorted(counts.items(), key=lambda item: item[1], reverse=True)
            if not sorted_counts:
                return {'type': 'string', 'format': 'default'}
            # Allow also counts that are not the max, based on the confidence
            max_count = sorted_counts[0][1]
            sorted_counts = filter(lambda item: item[1] >= max_count * confidence,
                                   sorted_counts)
            # Choose the most specific data type
            sorted_counts = sorted(sorted_counts,
                                   key=lambda item: item[0][2])
            rv = {'type': sorted_counts[0][0][0], 'format': sorted_counts[0][0][1]}
        return rv



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from . import config
__version__ = config.VERSION


# Module API

from .cli import cli
from .table import Table
from .schema import Schema
from .field import Field
from .storage import Storage
from .validate import validate
from .infer import infer
from .schema import FailedCast
from .exceptions import DataPackageException
from .exceptions import TableSchemaException
from .exceptions import LoadError
from .exceptions import ValidationError
from .exceptions import CastError
from .exceptions import IntegrityError
from .exceptions import UniqueKeyError
from .exceptions import RelationError
from .exceptions import UnresolvedFKError
from .exceptions import StorageError

# Deprecated

from . import exceptions



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
from ..config import ERROR


# Module API

def cast_year(format, value, **options):
    if not isinstance(value, int):
        if not isinstance(value, six.string_types):
            return ERROR
        if len(value) != 4:
            return ERROR
        try:
            value = int(value)
        except Exception:
            return ERROR
    if value < 0 or value > 9999:
        return ERROR
    return value



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


# Module API

class DataPackageException(Exception):
    """Base class for all DataPackage/TableSchema exceptions.

    If there are multiple errors, they can be read from the exception object:

    ```python
    try:
        # lib action
    except DataPackageException as exception:
        if exception.multiple:
            for error in exception.errors:
                # handle error
    ```

    """

    # Public

    def __init__(self, message, errors=None):
        self.__errors = errors or []
        super(Exception, self).__init__(message)

    @property
    def multiple(self):
        """Whether it's a nested exception

        # Returns
            bool: whether it's a nested exception

        """
        return bool(self.__errors)

    @property
    def errors(self):
        """List of nested errors

        # Returns
            DataPackageException[]: list of nested errors

        """
        return self.__errors


class TableSchemaException(DataPackageException):
    """Base class for all TableSchema exceptions.
    """
    pass


class LoadError(TableSchemaException):
    """All loading errors.
    """
    pass


class ValidationError(TableSchemaException):
    """All validation errors.
    """
    pass


class CastError(TableSchemaException):
    """All value cast errors.
    """
    pass


class IntegrityError(TableSchemaException):
    """All integrity errors.
    """
    pass


class UniqueKeyError(CastError):
    """Unique key constraint violation (CastError subclass)
    """
    pass


class RelationError(TableSchemaException):
    """All relations errors.
    """
    pass


class UnresolvedFKError(RelationError):
    """Unresolved foreign key reference error (RelationError subclass).
    """
    pass


class StorageError(TableSchemaException):
    """All storage errors.
    """
    pass


# Deprecated

MultipleInvalid = TableSchemaException
InvalidJSONError = LoadError
SchemaValidationError = ValidationError
InvalidSchemaError = ValidationError
InvalidCastError = CastError
ConstraintError = CastError



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


# Module API

from .enum import check_enum
from .maximum import check_maximum
from .maxLength import check_maxLength
from .minimum import check_minimum
from .minLength import check_minLength
from .pattern import check_pattern
from .required import check_required
from .unique import check_unique



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import decimal


# Module API

def check_minimum(constraint, value):
    if value is None:
        return True
    try:
        if value >= constraint:
            return True
    except decimal.InvalidOperation:
        # For non-finite numbers NaN, INF and -INF
        # the constraint always is not satisfied
        return False
    return False



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from datetime import time
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', time(6), time(6)),
    ('default', '06:00:00', time(6)),
    ('default', '09:00', ERROR),
    ('default', '3 am', ERROR),
    ('default', '3.00', ERROR),
    ('default', 'invalid', ERROR),
    ('default', True, ERROR),
    ('default', '', ERROR),
    ('any', time(6), time(6)),
    ('any', '06:00:00', time(6)),
    ('any', '3:00 am', time(3)),
    ('any', 'some night', ERROR),
    ('any', 'invalid', ERROR),
    ('any', True, ERROR),
    ('any', '', ERROR),
    ('%H:%M', time(6), time(6)),
    ('%H:%M', '06:00', time(6)),
    ('%M:%H', '06:50', ERROR),
    ('%H:%M', '3:00 am', ERROR),
    ('%H:%M', 'some night', ERROR),
    ('%H:%M', 'invalid', ERROR),
    ('%H:%M', True, ERROR),
    ('%H:%M', '', ERROR),
    ('invalid', '', ERROR),
    # Deprecated
    ('fmt:%H:%M', time(6), time(6)),
    ('fmt:%H:%M', '06:00', time(6)),
    ('fmt:%M:%H', '06:50', ERROR),
    ('fmt:%H:%M', '3:00 am', ERROR),
    ('fmt:%H:%M', 'some night', ERROR),
    ('fmt:%H:%M', 'invalid', ERROR),
    ('fmt:%H:%M', True, ERROR),
    ('fmt:%H:%M', '', ERROR),
])
def test_cast_time(format, value, result):
    with pytest.warns(None) as recorded:
        assert types.cast_time(format, value) == result
    if not format.startswith('fmt:'):
        assert recorded.list == []



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from datetime import date, datetime
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', date(2019, 1, 1), date(2019, 1, 1)),
    ('default', '2019-01-01', date(2019, 1, 1)),
    ('default', '10th Jan 1969', ERROR),
    ('default', 'invalid', ERROR),
    ('default', True, ERROR),
    ('default', '', ERROR),
    ('default', datetime(2018, 1, 1), date(2018, 1, 1)),
    ('default', datetime(2018, 3, 1, 8, 30, 23), ERROR),
    ('any', date(2019, 1, 1), date(2019, 1, 1)),
    ('any', '2019-01-01', date(2019, 1, 1)),
    ('any', '10th Jan 1969', date(1969, 1, 10)),
    ('any', '10th Jan nineteen sixty nine', ERROR),
    ('any', 'invalid', ERROR),
    ('any', True, ERROR),
    ('any', '', ERROR),
    ('%d/%m/%y', date(2019, 1, 1), date(2019, 1, 1)),
    ('%d/%m/%y', '21/11/06', date(2006, 11, 21)),
    ('%y/%m/%d', '21/11/06 16:30', ERROR),
    ('%d/%m/%y', 'invalid', ERROR),
    ('%d/%m/%y', True, ERROR),
    ('%d/%m/%y', '', ERROR),
    ('invalid', '21/11/06 16:30', ERROR),
    # Deprecated
    ('fmt:%d/%m/%y', date(2019, 1, 1), date(2019, 1, 1)),
    ('fmt:%d/%m/%y', '21/11/06', date(2006, 11, 21)),
    ('fmt:%y/%m/%d', '21/11/06 16:30', ERROR),
    ('fmt:%d/%m/%y', 'invalid', ERROR),
    ('fmt:%d/%m/%y', True, ERROR),
    ('fmt:%d/%m/%y', '', ERROR),
])
def test_cast_date(format, value, result):
    with pytest.warns(None) as recorded:
        assert types.cast_date(format, value) == result
    if not format.startswith('fmt:'):
        assert recorded.list == []



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from decimal import Decimal

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result, options', [
    ('default', 1, 1, {}),
    ('default', 1 << 63, 1 << 63, {}),
    ('default', '1', 1, {}),
    ('default', 1.0, 1, {}),
    ('default', Decimal('1.0'), 1, {}),
    ('default', '1$', 1, {'bareNumber': False}),
    ('default', 'ab1$', 1, {'bareNumber': False}),
    ('default', True, ERROR, {}),
    ('default', False, ERROR, {}),
    ('default', 3.14, ERROR, {}),
    ('default', '3.14', ERROR, {}),
    ('default', Decimal('3.14'), ERROR, {}),
    ('default', '', ERROR, {}),
])
def test_cast_integer(format, value, result, options):
    assert types.cast_integer(format, value, **options) == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import sys
import json
import click
import tableschema
from . import config


# Module API

@click.group(help='')
def cli():
    """Command-line interface

    ```
    Usage: tableschema [OPTIONS] COMMAND [ARGS]...

    Options:
      --help  Show this message and exit.

    Commands:
      infer     Infer a schema from data.
      info      Return info on this version of Table Schema
      validate  Validate that a supposed schema is in fact a Table Schema.
    ```

    """
    pass


@cli.command()
def info():
    """Return info on this version of Table Schema"""
    click.echo(json.dumps({'version': config.VERSION}, ensure_ascii=False, indent=4))


@cli.command()
@click.argument('data')
@click.option('--row_limit', default=100, type=int)
@click.option('--confidence', default=0.75, type=float)
@click.option('--encoding', default='utf-8')
@click.option('--to_file')
def infer(data, row_limit, confidence, encoding, to_file):
    """Infer a schema from data.

    - data must be a local filepath
    - data must be CSV
    - the file encoding is assumed to be UTF-8 unless an encoding is passed
      with --encoding
    - the first line of data must be headers
    - these constraints are just for the CLI

    """
    descriptor = tableschema.infer(data,
                                   encoding=encoding,
                                   limit=row_limit,
                                   confidence=confidence)
    if to_file:
        with io.open(to_file, mode='w+t', encoding='utf-8') as dest:
            dest.write(json.dumps(descriptor, ensure_ascii=False, indent=4))
    click.echo(descriptor)


@cli.command()
@click.argument('schema')
def validate(schema):
    """Validate that a supposed schema is in fact a Table Schema."""
    try:
        tableschema.validate(schema)
        click.echo("Schema is valid")
        sys.exit(0)
    except tableschema.exceptions.ValidationError as exception:
        click.echo("Schema is not valid")
        click.echo(exception.errors)
        sys.exit(1)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
from functools import partial
from cached_property import cached_property
from .profile import Profile
from . import constraints
from . import exceptions
from . import helpers
from . import config
from . import types


# Module API

class Field(object):
    """Field representaion

    # Arguments
        descriptor (dict): schema field descriptor
        missingValues (str[]): an array with string representing missing values

    # Raises
        TableSchemaException: raises any error that occurs during the process

    """

    # Public

    ERROR = config.ERROR

    def __init__(self, descriptor, missing_values=config.DEFAULT_MISSING_VALUES,
                 # Internal
                 schema=None):

        # Process descriptor
        descriptor = helpers.expand_field_descriptor(descriptor)

        # Set attributes
        self.__descriptor = descriptor
        self.__missing_values = missing_values
        self.__schema = schema
        self.__cast_function = self.__get_cast_function()
        self.__check_functions = self.__get_check_functions()

    @cached_property
    def schema(self):
        """Returns a schema instance if the field belongs to some schema

        # Returns
            Schema: field's schema

        """
        return self.__schema

    @cached_property
    def name(self):
        """Field name

        # Returns
            str: field name

        """
        return self.__descriptor.get('name')

    @cached_property
    def type(self):
        """Field type

        # Returns
            str: field type

        """
        return self.__descriptor.get('type')

    @cached_property
    def format(self):
        """Field format

        # Returns
            str: field format

        """
        return self.__descriptor.get('format')

    @cached_property
    def missing_values(self):
        """Field's missing values

        # Returns
            str[]: missing values

        """
        return self.__missing_values

    @cached_property
    def required(self):
        """Whether field is required

        # Returns
            bool: true if required

        """
        return self.constraints.get('required', False)

    @cached_property
    def constraints(self):
        """Field constraints

        # Returns
            dict: dict of field constraints

        """
        return self.__descriptor.get('constraints', {})

    @cached_property
    def descriptor(self):
        """Fields's descriptor

        # Returns
            dict: descriptor

        """
        return self.__descriptor

    @cached_property
    def cast_function(self):
        return self.__cast_function

    @cached_property
    def check_functions(self):
        return self.__check_functions

    def cast_value(self, value, constraints=True):
        """Cast given value according to the field type and format.

        # Arguments
            value (any): value to cast against field
            constraints (boll/str[]): gets constraints configuration
                - it could be set to true to disable constraint checks
                - it could be an Array of constraints to check e.g. ['minimum', 'maximum']

        # Raises
            TableSchemaException: raises any error that occurs during the process

        # Returns
            any: returns cast value

        """

        # Null value
        if value in self.__missing_values:
            # Whether missing_values should be preserved without being cast
            if os.environ.get('TABLESCHEMA_PRESERVE_MISSING_VALUES'):
                return value
            value = None

        # Cast value
        cast_value = value
        if value is not None:
            cast_value = self.__cast_function(value)
            if cast_value == config.ERROR:
                raise exceptions.CastError((
                    'Field "{field.name}" can\'t cast value "{value}" '
                    'for type "{field.type}" with format "{field.format}"'
                    ).format(field=self, value=value))

        # Check value
        if constraints:
            for name, check in self.__check_functions.items():
                if isinstance(constraints, list):
                    if name not in constraints:
                        continue
                passed = check(cast_value)
                if not passed:
                    raise exceptions.CastError((
                        'Field "{field.name}" has constraint "{name}" '
                        'which is not satisfied for value "{value}"'
                        ).format(field=self, name=name, value=value))

        return cast_value

    def test_value(self, value, constraints=True):
        """Test whether value is compliant to the field.

        # Arguments
            value (any): value to cast against field
            constraints (bool/str[]): constraints configuration

        # Returns
            bool: returns if value is compliant to the field

        """
        try:
            self.cast_value(value, constraints=constraints)
        except exceptions.CastError:
            return False
        return True

    # Private

    def __get_cast_function(self):
        options = {}
        # Get cast options
        for key in ['decimalChar', 'groupChar', 'bareNumber', 'trueValues', 'falseValues']:
            value = self.descriptor.get(key)
            if value is not None:
                options[key] = value
        try:
            cast = getattr(types, 'cast_%s' % self.type)
        except AttributeError:
            message = 'Not supported field type: %s' % self.type
            raise exceptions.TableSchemaException(message)
        cast = partial(cast, self.format, **options)
        return cast

    def __get_check_functions(self):
        checks = {}
        cast = partial(self.cast_value, constraints=False)
        whitelist = _get_field_constraints(self.type)
        for name, constraint in self.constraints.items():
            if name in whitelist:
                # Cast enum constraint
                if name in ['enum']:
                    constraint = list(map(cast, constraint))
                # Cast maximum/minimum constraint
                if name in ['maximum', 'minimum']:
                    constraint = cast(constraint)
                check = getattr(constraints, 'check_%s' % name)
                checks[name] = partial(check, constraint)
        return checks


# Internal

def _get_field_constraints(type):
    # Extract list of constraints for given type from jsonschema
    jsonschema = Profile('table-schema').jsonschema
    profile_types = jsonschema['properties']['fields']['items']['anyOf']
    for profile_type in profile_types:
        if type in profile_type['properties']['type']['enum']:
            return profile_type['properties']['constraints']['properties'].keys()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from tabulator import Stream
from functools import partial
from collections import OrderedDict
from .storage import Storage
from .schema import Schema
from . import exceptions
from . import helpers
from . import config
from collections import defaultdict


# Module API

class Table(object):
    """Table representation

    # Arguments
      source (str/list[]): data source one of:
        - local file (path)
        - remote file (url)
        - array of arrays representing the rows
      schema (any): data schema in all forms supported by `Schema` class
      strict (bool): strictness option to pass to `Schema` constructor
      post_cast (function[]): list of post cast processors
      storage (None): storage name like `sql` or `bigquery`
      options (dict): `tabulator` or storage's options

    # Raises
      TableSchemaException: raises on any error

    """

    # Public

    def __init__(self, source, schema=None, strict=False,
                 post_cast=[], storage=None, **options):

        # Set attributes
        self.__source = source
        self.__stream = None
        self.__schema = None
        self.__headers = None
        self.__storage = None
        self.__post_cast = copy(post_cast)

        # Schema
        if isinstance(schema, Schema):
            self.__schema = schema
        elif schema is not None:
            self.__schema = Schema(schema)

        # Stream (tabulator)
        if storage is None:
            options.setdefault('headers', 1)
            self.__stream = Stream(source,  **options)

        # Stream (storage)
        else:
            if not isinstance(storage, Storage):
                storage = Storage.connect(storage, **options)
            if self.__schema:
                storage.describe(source, self.__schema.descriptor)
            headers = Schema(storage.describe(source)).field_names
            self.__stream = Stream(partial(storage.iter, source), headers=headers)
            self.__storage = storage

    @property
    def headers(self):
        """Table's headers is available

        # Returns
            str[]: headers

        """
        return self.__headers

    @property
    def schema(self):
        """Returns schema class instance if available

        # Returns
            Schema: schema

        """
        return self.__schema

    @property
    def size(self):
        """Table's size in BYTES if it's available

        If it's already read using e.g. `table.read`, otherwise returns `None`.
        In the middle of an iteration it returns size of already read contents

        # Returns
            int/None: size in BYTES

        """
        if self.__stream:
            return self.__stream.size

    @property
    def hash(self):
        """Table's SHA256 hash if it's available.

        If it's already read using e.g. `table.read`, otherwise returns `None`.
        In the middle of an iteration it returns hash of already read contents

        # Returns
            str/None: SHA256 hash

        """
        if self.__stream:
            return self.__stream.hash

    def iter(self, keyed=False, extended=False, cast=True,
             integrity=False, relations=False,
             foreign_keys_values=False, exc_handler=None):
        """Iterates through the table data and emits rows cast based on table schema.

        # Arguments

            keyed (bool):
                yield keyed rows in a form of `{header1\\: value1, header2\\: value2}`
                (default is false; the form of rows is `[value1, value2]`)

            extended (bool):
                yield extended rows in a for of `[rowNumber, [header1, header2], [value1, value2]]`
                (default is false; the form of rows is `[value1, value2]`)

            cast (bool):
                disable data casting if false
                (default is true)

            integrity (dict):
                dictionary in a form of `{'size'\\: <bytes>, 'hash'\\: '<sha256>'}`
                to check integrity of the table when it's read completely.
                Both keys are optional.

            relations (dict):
                dictionary of foreign key references in a form
                of `{resource1\\: [{field1\\: value1, field2\\: value2}, ...], ...}`.
                If provided, foreign key fields will checked and resolved
                to one of their references (/!\\ one-to-many fk are not completely resolved).

            foreign_keys_values (dict):
                three-level dictionary of foreign key references optimized
                to speed up validation process in a form of
                `{resource1\\: {(fk_field1, fk_field2)\\: {(value1, value2)\\: {one_keyedrow}, ... }}}`.
                If not provided but relations is true, it will be created
                before the validation process by *index_foreign_keys_values* method

            exc_handler (func):
                optional custom exception handler callable.
                Can be used to defer raising errors (i.e. "fail late"), e.g.
                for data validation purposes. Must support the signature below

        # Custom exception handler

        ```python
        def exc_handler(exc, row_number=None, row_data=None, error_data=None):
            '''Custom exception handler (example)

            # Arguments:
                exc(Exception):
                    Deferred exception instance
                row_number(int):
                    Data row number that triggers exception exc
                row_data(OrderedDict):
                    Invalid data row source data
                error_data(OrderedDict):
                    Data row source data field subset responsible for the error, if
                    applicable (e.g. invalid primary or foreign key fields). May be
                    identical to row_data.
            '''
            # ...
        ```

        # Raises
            TableSchemaException: base class of any error
            CastError: data cast error
            IntegrityError: integrity checking error
            UniqueKeyError: unique key constraint violation
            UnresolvedFKError: unresolved foreign key reference error

        # Returns
            Iterator[list]: yields rows

        """
        exc_handler = helpers.default_exc_handler if exc_handler is None else \
            exc_handler

        # Prepare unique checks
        if cast:
            unique_fields_cache = {}
            if self.schema:
                unique_fields_cache = _create_unique_fields_cache(self.schema)
        # Prepare relation checks
        if relations and not foreign_keys_values:
            # we have to test relations but the index has not been precomputed
            # prepare the index to boost validation process
            foreign_keys_values = self.index_foreign_keys_values(relations)

        # Open/iterate stream
        with self.__stream as stream:
            iterator = stream.iter(extended=True)
            iterator = self.__apply_processors(
                iterator, cast=cast, exc_handler=exc_handler)
            for row_number, headers, row in iterator:

                # Get headers
                if not self.__headers:
                    self.__headers = headers

                # Check headers
                if cast:
                    if self.schema and self.headers:
                        if self.headers != self.schema.field_names:
                            message = (
                                'Table headers (%r) don\'t match '
                                'schema field names (%r) in row %s' % (
                                    self.headers, self.schema.field_names,
                                    row_number))
                            keyed_row = OrderedDict(zip(headers, row))
                            exc_handler(
                                exceptions.CastError(message),
                                row_number=row_number, row_data=keyed_row,
                                error_data=keyed_row)
                            continue

                # Check unique
                if cast:
                    for indexes, cache in unique_fields_cache.items():
                        keyed_values = OrderedDict(
                            (headers[i], value)
                            for i, value in enumerate(row) if i in indexes)
                        values = tuple(keyed_values.values())
                        if not all(map(lambda value: value is None, values)):
                            if values in cache['data']:
                                message = (
                                    'Field(s) "%s" duplicates in row "%s" '
                                    'for values %r' % (
                                        cache['name'], row_number, values))
                                exc_handler(
                                    exceptions.UniqueKeyError(message),
                                    row_number=row_number,
                                    row_data=OrderedDict(zip(headers, row)),
                                    error_data=keyed_values)
                            cache['data'].add(values)

                # Resolve relations
                if relations:
                    if self.schema:
                        row_with_relations = dict(zip(headers, copy(row)))
                        for foreign_key in self.schema.foreign_keys:
                            refValue = _resolve_relations(row, headers, foreign_keys_values,
                                                          foreign_key)
                            if refValue is None:
                                keyed_row = OrderedDict(zip(headers, row))
                                # local values of the FK
                                local_keyed_values = {
                                    key: keyed_row[key]
                                    for key in foreign_key['fields']
                                    }
                                local_values = tuple(local_keyed_values.values())
                                message = (
                                    'Foreign key "%s" violation in row "%s": '
                                    '%s not found in %s' % (
                                        foreign_key['fields'],
                                        row_number,
                                        local_values,
                                        foreign_key['reference']['resource']))
                                exc_handler(
                                    exceptions.UnresolvedFKError(message),
                                    row_number=row_number, row_data=keyed_row,
                                    error_data=local_keyed_values)
                                # If we reach this point we don't fail-early
                                # i.e. no exception has been raised. As the
                                # reference can't be resolved, use empty dict
                                # as the "unresolved result".
                                for field in foreign_key['fields']:
                                    if not isinstance(
                                            row_with_relations[field], dict):
                                        row_with_relations[field] = {}
                            elif type(refValue) is dict:
                                # Substitute resolved referenced object for
                                # original referencing field value.
                                # For a composite foreign key, this substitutes
                                # each part of the composite key with the
                                # referenced object.
                                for field in foreign_key['fields']:
                                    if type(row_with_relations[field]) is not dict:
                                        # no previous refValues injected on this field
                                        row_with_relations[field] = refValue
                                    else:
                                        # alreayd one ref, merging
                                        row_with_relations[field].update(refValue)
                            else:
                                # case when all original value of the FK are empty
                                # refValue == row, there is nothing to do
                                # an empty dict might be a better returned value for this case ?
                                pass

                        #  mutate row now that we are done, in the right order
                        row = [row_with_relations[f] for f in headers]

                # Form row
                if extended:
                    yield (row_number, headers, row)
                elif keyed:
                    yield dict(zip(headers, row))
                else:
                    yield row

            # Check integrity
            if integrity:
                violations = []
                size = integrity.get('size')
                hash = integrity.get('hash')
                if size and size != self.__stream.size:
                    violations.append('size "%s"' % self.__stream.size)
                if hash and hash != self.__stream.hash:
                    violations.append('hash "%s"' % self.__stream.hash)
                if violations:
                    message = 'Calculated %s differ(s) from declared value(s)'
                    raise exceptions.IntegrityError(message % ' and '.join(violations))

    def read(self, keyed=False, extended=False, cast=True, limit=None,
             integrity=False, relations=False, foreign_keys_values=False,
             exc_handler=None):
        """Read the whole table and return as array of rows

        > It has the same API as `table.iter` except for

        # Arguments
            limit (int): limit count of rows to read and return

        # Returns
            list[]: returns rows

        """
        result = []
        rows = self.iter(
            keyed=keyed, extended=extended, cast=cast, integrity=integrity,
            relations=relations, foreign_keys_values=foreign_keys_values,
            exc_handler=exc_handler)
        for count, row in enumerate(rows, start=1):
            result.append(row)
            if count == limit:
                break
        return result

    def infer(self, limit=100, confidence=0.75,
              missing_values=config.DEFAULT_MISSING_VALUES,
              guesser_cls=None, resolver_cls=None):
        """Infer a schema for the table.

        It will infer and set Table Schema to `table.schema` based on table data.

        # Arguments
            limit (int): limit rows sample size
            confidence (float): how many casting errors are allowed (as a ratio, between 0 and 1)
            missing_values (str[]): list of missing values (by default `['']`)
            guesser_cls (class): you can implement inferring strategies by
                 providing type-guessing and type-resolving classes [experimental]
            resolver_cls (class): you can implement inferring strategies by
                 providing type-guessing and type-resolving classes [experimental]

        # Returns
            dict: Table Schema descriptor

        """
        if self.__schema is None or self.__headers is None:

            # Infer (tabulator)
            if not self.__storage:
                with self.__stream as stream:
                    if self.__schema is None:
                        self.__schema = Schema({'missingValues': missing_values})
                        self.__schema.infer(stream.sample[:limit],
                                            headers=stream.headers,
                                            confidence=confidence,
                                            guesser_cls=guesser_cls,
                                            resolver_cls=resolver_cls)
                    if self.__headers is None:
                        self.__headers = stream.headers

            # Infer (storage)
            else:
                descriptor = self.__storage.describe(self.__source)
                if self.__schema is None:
                    self.__schema = Schema(descriptor)
                if self.__headers is None:
                    self.__headers = self.__schema.field_names

        return self.__schema.descriptor

    def save(self, target, storage=None, **options):
        """Save data source to file locally in CSV format with `,` (comma) delimiter

        > To save schema use `table.schema.save()`

        # Arguments
            target (str): saving target (e.g. file path)
            storage (None/str): storage name like `sql` or `bigquery`
            options (dict): `tabulator` or storage options

        # Raises
            TableSchemaException: raises an error if there is saving problem

        # Returns
            True/Storage: returns true or storage instance

        """

        # Save (tabulator)
        if storage is None:
            with Stream(self.iter, headers=self.__schema.headers) as stream:
                stream.save(target, **options)
            return True

        # Save (storage)
        else:
            if not isinstance(storage, Storage):
                storage = Storage.connect(storage, **options)
            storage.create(target, self.__schema.descriptor, force=True)
            storage.write(target, self.iter(cast=False))
            return storage

    def index_foreign_keys_values(self, relations):
        """Creates a three-level dictionary of foreign key references

        We create them optimized to speed up validation process in a form of
        `{resource1: {(fk_field1, fk_field2): {(value1, value2): {one_keyedrow}, ... }}}`.

        For each foreign key of the schema it will iterate through the corresponding
        `relations['resource']` to create an index (i.e. a dict) of existing values
        for the foreign fields and store on keyed row for each value combination.

        The optimization relies on the indexation of possible values for one foreign key
        in a hashmap to later speed up resolution.

        This method is public to allow creating the index once to apply it
        on multiple tables charing the same schema
        (typically [grouped resources in datapackage](https://github.com/frictionlessdata/datapackage-py#group))

        # Notes

        - the second key of the output is a tuple of the foreign fields,
            a proxy identifier of the foreign key
        - the same relation resource can be indexed multiple times
            as a schema can contain more than one Foreign Keys
            pointing to the same resource

        # Arguments
            relations (dict):
                dict of foreign key references in a form of
                `{resource1\\: [{field1\\: value1, field2\\: value2}, ...], ...}`.
                It must contain all resources pointed in the foreign keys schema definition.

        # Returns
            dict:
                returns a three-level dictionary of foreign key references
                optimized to speed up validation process in a form of
                `{resource1\\: {(fk_field1, fk_field2)\\: {(value1, value2)\\: {one_keyedrow}, ... }}})`

        """

        # we dont need to load the complete reference table to test relations
        # we can lower payload AND optimize testing foreign keys
        # by preparing the right index based on the foreign key definition
        # foreign_keys are sets of tuples of all possible values in the foreign table
        # foreign keys =
        # [reference] [foreign_keys tuple] = { (foreign_keys_values, ) : one_keyedrow, ... }
        foreign_keys = defaultdict(dict)
        if self.schema:
            for fk in self.schema.foreign_keys:
                # load relation data
                relation = fk['reference']['resource']

                # create a set of foreign keys
                # to optimize we prepare index of existing values
                # this index should use reference + foreign_keys as key
                # cause many foreign keys may use the same reference
                foreign_keys[relation][tuple(fk['reference']['fields'])] = {}
                for row in relations[relation]:
                    key = tuple([row[foreign_field] for foreign_field in fk['reference']['fields']])
                    # here we should chose to pick the first or nth row which match
                    # previous implementation picked the first, so be it
                    if key not in foreign_keys[relation][tuple(fk['reference']['fields'])]:
                        foreign_keys[relation][tuple(fk['reference']['fields'])][key] = row
        return foreign_keys

    # Private

    def __apply_processors(self, iterator, cast=True, exc_handler=None):

        # Apply processors to iterator
        def builtin_processor(extended_rows):
            for row_number, headers, row in extended_rows:
                if self.__schema and cast:
                    row = self.__schema.cast_row(
                        row, row_number=row_number, exc_handler=exc_handler)
                yield (row_number, headers, row)
        processors = [builtin_processor] + self.__post_cast
        for processor in processors:
            iterator = processor(iterator)

        return iterator


# Internal

def _create_unique_fields_cache(schema):
    primary_key_indexes = []
    cache = {}

    # Unique
    for index, field in enumerate(schema.fields):
        if field.name in schema.primary_key:
            primary_key_indexes.append(index)
        if field.constraints.get('unique'):
            cache[tuple([index])] = {
                'name': field.name,
                'data': set(),
            }

    # Primary key
    if primary_key_indexes:
        cache[tuple(primary_key_indexes)] = {
            'name': ', '.join(schema.primary_key),
            'data': set(),
        }

    return cache


def _resolve_relations(row, headers, foreign_keys_values, foreign_key):

    # Prepare helpers - needed data structures
    keyed_row = OrderedDict(zip(headers, row))
    # local values of the FK
    local_values = tuple(keyed_row[f] for f in foreign_key['fields'])
    if set(local_values) != {None}:
        # test existence into the foreign
        relation = foreign_key['reference']['resource']
        keys = tuple(foreign_key['reference']['fields'])
        foreign_values = foreign_keys_values[relation][keys]
        if local_values in foreign_values:
            return foreign_values[local_values]
        else:
            return None
    else:
        # empty values for all keys, return original values
        return row



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from decimal import Decimal
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result, options', [
    ('default', Decimal(1), Decimal(1), {}),
    ('default', 1, Decimal(1), {}),
    ('default', 1.0, Decimal(1), {}),
    ('default', 1 << 63, Decimal(1 << 63), {}),
    ('default', '1', Decimal(1), {}),
    ('default', '10.00', Decimal(10), {}),
    ('default', '10.50', Decimal(10.5), {}),
    ('default', 24.122667, Decimal('24.122667'), {}),
    ('default', '100%', Decimal(100), {'bareNumber': False}),
    ('default', '1000‰', Decimal(1000), {'bareNumber': False}),
    ('default', '-1000', Decimal(-1000), {}),
    ('default', '1,000', Decimal(1000), {'groupChar': ','}),
    ('default', '10,000.00', Decimal(10000), {'groupChar': ','}),
    ('default', '10,000,000.50', Decimal(10000000.5), {'groupChar': ','}),
    ('default', '10#000.00', Decimal(10000), {'groupChar': '#'}),
    ('default', '10#000#000.50', Decimal(10000000.5), {'groupChar': '#'}),
    ('default', '10.50', Decimal(10.5), {'groupChar': '#'}),
    ('default', '1#000', Decimal(1000), {'groupChar': '#'}),
    ('default', '10#000@00', Decimal(10000), {'groupChar': '#', 'decimalChar': '@'}),
    ('default', '10#000#000@50', Decimal(10000000.5), {'groupChar': '#', 'decimalChar': '@'}),
    ('default', '10@50', Decimal(10.5), {'groupChar': '#', 'decimalChar': '@'}),
    ('default', '1#000', Decimal(1000), {'groupChar': '#', 'decimalChar': '@'}),
    ('default', '10,000.00', Decimal(10000), {'groupChar': ',', 'bareNumber': False}),
    ('default', '10,000,000.00', Decimal(10000000), {'groupChar': ',', 'bareNumber': False}),
    ('default', '10.000.000,00', Decimal(10000000), {'groupChar': '.', 'decimalChar': ','}),
    ('default', '$10000.00', Decimal(10000), {'bareNumber': False}),
    ('default', '  10,000.00 €', Decimal(10000), {'groupChar': ',', 'bareNumber': False}),
    ('default', '10 000,00', Decimal(10000), {'groupChar': ' ', 'decimalChar': ','}),
    ('default', '10 000 000,00', Decimal(10000000), {'groupChar': ' ', 'decimalChar': ','}),
    ('default', '10000,00 ₪', Decimal(10000), {'groupChar': ' ', 'decimalChar': ',', 'bareNumber': False}),
    ('default', '  10 000,00 £', Decimal(10000), {'groupChar': ' ', 'decimalChar': ',', 'bareNumber': False}),
    ('default', True, ERROR, {}),
    ('default', False, ERROR, {}),
    ('default', '10,000a.00', ERROR, {}),
    ('default', '10+000.00', ERROR, {}),
    ('default', '$10:000.00', ERROR, {}),
    ('default', 'string', ERROR, {}),
    ('default', '', ERROR, {}),
])
def test_cast_number(format, value, result, options):
    assert types.cast_number(format, value, **options) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from six import add_metaclass
from importlib import import_module
from abc import ABCMeta, abstractmethod
from . import exceptions


# Module API

@add_metaclass(ABCMeta)
class Storage(object):
    """Storage factory/interface

    # For users

    > Use `Storage.connect` to instantiate a storage

    For instantiation of concrete storage instances,
    `tableschema.Storage` provides a unified factory method `connect`
    (which uses the plugin system under the hood):

    ```python
    # pip install tableschema_sql
    from tableschema import Storage

    storage = Storage.connect('sql', **options)
    storage.create('bucket', descriptor)
    storage.write('bucket', rows)
    storage.read('bucket')
    ```

    # For integrators

    The library includes interface declaration to implement tabular `Storage`.
    This interface allow to use different data storage systems like SQL
    with `tableschema.Table` class (load/save) as well as on the data package level:

    ![Storage](https://raw.githubusercontent.com/frictionlessdata/tableschema-py/master/data/storage.png)

    An implementor must follow `tableschema.Storage` interface
    to write his own storage backend. Concrete storage backends
    could include additional functionality specific to conrete storage system.
    See `plugins` below to know how to integrate custom storage plugin into your workflow.

    """

    # Public

    @abstractmethod
    def __init__(self, **options):
        pass

    @classmethod
    def connect(cls, name, **options):
        """Create tabular `storage` based on storage name.

        > This method is statis: `Storage.connect()`

        # Arguments
            name (str): storage name like `sql`
            options (dict): concrete storage options

        # Raises
            StorageError: raises on any error

        # Returns
            Storage: returns `Storage` instance

        """
        if cls is not Storage:
            message = 'Storage.connect is not available on concrete implemetations'
            raise exceptions.StorageError(message)
        module = 'tableschema.plugins.%s' % name
        storage = import_module(module).Storage(**options)
        return storage

    @property
    @abstractmethod
    def buckets(self):
        """Return list of storage bucket names.

        A `bucket` is a special term which has almost the same meaning as `table`.
        You should consider `bucket` as a `table` stored in the `storage`.

        # Raises
            exceptions.StorageError: raises on any error

        # Returns
            str[]: return list of bucket names

        """
        pass

    @abstractmethod
    def create(self, bucket, descriptor, force=False):
        """Create one/multiple buckets.

        # Arguments
            bucket (str/list): bucket name or list of bucket names
            descriptor (dict/dict[]): schema descriptor or list of descriptors
            force (bool): whether to delete and re-create already existing buckets

        # Raises
            exceptions.StorageError: raises on any error

        """
        pass

    @abstractmethod
    def delete(self, bucket=None, ignore=False):
        """ Delete one/multiple/all buckets.

        # Arguments
            bucket (str/list/None): bucket name or list of bucket names to delete.
                If `None`, all buckets will be deleted
            descriptor (dict/dict[]): schema descriptor or list of descriptors
            ignore (bool): don't raise an error on non-existent bucket deletion

        # Raises
            exceptions.StorageError: raises on any error

        """
        pass

    @abstractmethod
    def describe(self, bucket, descriptor=None):
        """ Get/set bucket's Table Schema descriptor

        # Arguments
            bucket (str): bucket name
            descriptor (dict/None): schema descriptor to set

        # Raises
            exceptions.StorageError: raises on any error

        # Returns
            dict: returns Table Schema descriptor

        """
        pass

    @abstractmethod
    def iter(self, bucket):
        """Return an iterator of typed values based on the schema of this bucket.

        # Arguments
            bucket (str): bucket name

        # Raises
            exceptions.StorageError: raises on any error

        # Returns
            list[]: yields data rows

        """
        pass

    @abstractmethod
    def read(self, bucket):
        """Read typed values based on the schema of this bucket.

        # Arguments
            bucket (str): bucket name
        # Raises
            exceptions.StorageError: raises on any error
        # Returns
            list[]: returns data rows

        """
        pass

    @abstractmethod
    def write(self, bucket, rows):
        """ This method writes data rows into `storage`.

        It should store values of unsupported types as strings internally (like csv does).

        # Arguments
            bucket (str): bucket name
            rows (list[]): data rows to write

        # Raises
            exceptions.StorageError: raises on any error

        """
        pass



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
from setuptools import setup, find_packages


# Helpers
def read(*paths):
    """Read a text file."""
    basedir = os.path.dirname(__file__)
    fullpath = os.path.join(basedir, *paths)
    contents = io.open(fullpath, encoding='utf-8').read().strip()
    return contents


# Prepare
PACKAGE = 'tableschema'
NAME = PACKAGE.replace('_', '-')
INSTALL_REQUIRES = [
    'six>=1.9',
    'click>=3.3',
    'requests>=2.5',
    'cached-property>=1.5',
    'python-dateutil>=2.4',
    # NOTE: Can be removed after a fix:
    # https://github.com/tobgu/pyrsistent/issues/208
    'pyrsistent<0.17',
    'jsonschema>=2.5',
    'unicodecsv>=0.14',
    'isodate>=0.5.4',
    'rfc3986>=1.1.0',
    'tabulator>=1.51.3',
]
TESTS_REQUIRE = [
    'mock',
    'pylama',
    'pytest',
    'pytest-cov',
    'tox',
]
README = read('README.md')
VERSION = read(PACKAGE, 'VERSION')
PACKAGES = find_packages(exclude=['examples', 'tests'])


# Run
setup(
    name=NAME,
    version=VERSION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={'develop': TESTS_REQUIRE},
    entry_points={
        'console_scripts': [
            'tableschema = tableschema.__main__:cli',
        ]
    },
    zip_safe=False,
    long_description=README,
    long_description_content_type="text/markdown",
    description='A utility library for working with Table Schema in Python',
    author='Open Knowledge Foundation',
    author_email='info@okfn.org',
    url='https://github.com/frictionlessdata/tableschema-py',
    license='MIT',
    keywords=[
        'frictionless data',
        'open data',
        'json schema',
        'table schema',
        'data package',
        'tabular data package',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import json
from ..config import ERROR
from ..profile import Profile


# Module API

def cast_geojson(format, value, **options):
    if isinstance(value, six.string_types):
        try:
            value = json.loads(value)
        except Exception:
            return ERROR
    if not isinstance(value, dict):
        return ERROR
    if format == 'default':
        try:
            _profile.validate(value)
        except Exception:
            return ERROR
    elif format == 'topojson':
        pass  # Accept any dict as possibly topojson for now
    return value


# Internal

_profile = Profile('geojson')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import json
from collections import namedtuple
from decimal import Decimal
from ..config import ERROR


# Module API

def cast_geopoint(format, value, **options):

    # Parse
    if isinstance(value, six.string_types):
        try:
            if format == 'default':
                lon, lat = value.split(',')
                lon = lon.strip()
                lat = lat.strip()
            elif format == 'array':
                lon, lat = json.loads(value)
            elif format == 'object':
                if isinstance(value, six.string_types):
                    value = json.loads(value)
                if len(value) != 2:
                    return ERROR
                lon = value['lon']
                lat = value['lat']
            value = _geopoint(Decimal(lon), Decimal(lat))
        except Exception:
            return ERROR

    # Validate
    try:
        value = _geopoint(*value)
        if value.lon > 180 or value.lon < -180:
            return ERROR
        if value.lat > 90 or value.lat < -90:
            return ERROR
    except Exception:
        return ERROR

    return value


# Internal

_geopoint = namedtuple('geopoint', ['lon', 'lat'])
_geopoint.__repr__ = lambda self: str([float(self[0]), float(self[1])])



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import warnings
from . import config
from .table import Table


# Module API

def infer(source, headers=1, limit=100, confidence=0.75,
          missing_values=config.DEFAULT_MISSING_VALUES,
          guesser_cls=None, resolver_cls=None,
          **options):
    """Infer source schema.

    # Arguments
        source (any): source as path, url or inline data
        headers (int/str[]): headers rows number or headers list
        confidence (float): how many casting errors are allowed (as a ratio, between 0 and 1)
        missing_values (str[]): list of missing values (by default `['']`)
        guesser_cls (class): you can implement inferring strategies by
            providing type-guessing and type-resolving classes [experimental]
        resolver_cls (class): you can implement inferring strategies by
            providing type-guessing and type-resolving classes [experimental]

    # Raises
        TableSchemaException: raises any error that occurs during the process

    # Returns
        dict: returns schema descriptor

    """

    # Deprecated arguments order
    is_string = lambda value: isinstance(value, six.string_types)
    if isinstance(source, list) and all(map(is_string, source)):
        warnings.warn('Correct arguments order infer(source, headers)', UserWarning)
        source, headers = headers, source

    table = Table(source, headers=headers, sample_size=limit, **options)
    descriptor = table.infer(limit=limit, confidence=confidence,
        missing_values=missing_values, guesser_cls=guesser_cls,
        resolver_cls=resolver_cls)
    return descriptor



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import six
from decimal import Decimal
from ..config import ERROR


# Module API

def cast_number(format, value, **options):
    group_char = options.get('groupChar', _DEFAULT_GROUP_CHAR)
    decimal_char = options.get('decimalChar', _DEFAULT_DECIMAL_CHAR)
    if not isinstance(value, Decimal):
        if isinstance(value, six.string_types):
            value = re.sub(r'\s', '', value)
            value = value.replace(decimal_char, '__decimal_char__')
            value = value.replace(group_char, '')
            value = value.replace('__decimal_char__', '.')
            if not options.get('bareNumber', _DEFAULT_BARE_NUMBER):
                value = re.sub(r'((^\D*)|(\D*$))', '', value)
        elif not isinstance(value, six.integer_types + (float,)):
            return ERROR
        elif value is True or value is False:
            return ERROR
        try:
            if isinstance(value, float):
                value = str(value)
            value = Decimal(value)
        except Exception:
            return ERROR
    return value


# Internal

_DEFAULT_GROUP_CHAR = ''
_DEFAULT_DECIMAL_CHAR = '.'
_DEFAULT_BARE_NUMBER = True
